package it.inaf.oats.ia2.tapschemamanager.businesslayer;

import it.inaf.oats.ia2.tapschemamanager.datalayer.Credentials;
import it.inaf.oats.ia2.tapschemamanager.datalayer.SchemaEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.TapSchemaHandler;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
public class TapSchema implements EntityWrapperContainer {

    private final TapSchemaHandler tapSchemaHandler;

    private final String name;

    private String selectedSchema;
    private final Map<String, Schema> schemas;

    public TapSchema(Credentials credentials, String tapSchemaName, boolean exists) throws SQLException {
        this.name = tapSchemaName;
        schemas = new HashMap<String, Schema>();

        this.tapSchemaHandler = new TapSchemaHandler(credentials, tapSchemaName, exists);

        if (exists) {
            for (SchemaEntity schemaEntity : tapSchemaHandler.getSchemas()) {
                Schema schema = new Schema(tapSchemaHandler, schemaEntity);
                schemas.put(schema.getName(), schema);
            }
            if (!schemas.isEmpty()) {
                selectEntity(schemas.keySet().iterator().next()); // select first schema
            }
        }
    }

    @Override
    public List<String> getAddables() {
        return Util.getAddables(tapSchemaHandler.getAllSchemas(), schemas.keySet());
    }

    @Override
    public void addEntityWrapper(String name) throws SQLException {
        Schema schema = schemas.get(name);
        SchemaEntity schemaEntity;
        if (schema == null) {
            // Add new
            schemaEntity = tapSchemaHandler.getNewSchema(name);
            schema = new Schema(tapSchemaHandler, schemaEntity);
            schemas.put(name, schema);
            schema.setStatus(Status.ADDED_NOT_PERSISTED);
        } else {
            schemaEntity = schema.getSchemaEntity();
            // Undo remove
            schema.setStatus(Status.ADDED_PERSISTED);
        }
        selectedSchema = name;

        tapSchemaHandler.addSchema(schemaEntity);
    }

    @Override
    public void removeEntityWrapper(String name) {
        Schema schema = schemas.get(name);
        if (schema.getStatus() == Status.ADDED_NOT_PERSISTED) {
            schemas.remove(name);
        } else {
            schema.setStatus(Status.TO_REMOVE);
        }
        selectedSchema = schemas.isEmpty() ? null : schemas.keySet().iterator().next();

        tapSchemaHandler.removeSchema(name);
    }

    public Schema getSchema(String name) {
        return schemas.get(name);
    }

    public List<String> getSchemasNames() {
        return Util.setToList(schemas.keySet());
    }

    public String getName() {
        return name;
    }

    public void save() throws SQLException {
        tapSchemaHandler.save();

        Iterator<Schema> iterator = schemas.values().iterator();
        while (iterator.hasNext()) {
            Schema schema = iterator.next();
            if (schema.getStatus() == Status.TO_REMOVE) {
                iterator.remove();
            } else {
                schema.setPersisted();
            }
        }
    }

    public Schema getSelectedSchema() {
        return selectedSchema == null ? null : schemas.get(selectedSchema);
    }

    public Table getSelectedTable() {
        Schema schema = getSelectedSchema();
        return schema == null ? null : schema.getSelectedTable();
    }

    public Column getSelectedColumn() {
        Table table = getSelectedTable();
        return table == null ? null : table.getSelectedColumn();
    }

    @Override
    public final void selectEntity(String entityName) {
        selectedSchema = entityName;
        Schema schema = schemas.get(entityName);
        List<String> tables = schema.getTablesNames();
        if (!tables.isEmpty()) {
            schema.selectEntity(tables.get(0));
        }
    }

    @Override
    public boolean hasChildren() {
        return !schemas.isEmpty();
    }

    @Override
    public List<EntityWrapper> getAllEntityWrappers() {
        return new ArrayList(schemas.values());
    }

    @Override
    public EntityWrapper getSelectedEntity() {
        if (selectedSchema == null) {
            return null;
        }
        return schemas.get(selectedSchema);
    }
}
