package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.persistence.EntityManager;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;

class ColumnsInfo {

    private final Map<String, Map<String, List<ColumnInfo>>> columnsInfo;

    ColumnsInfo() {
        this.columnsInfo = new HashMap<String, Map<String, List<ColumnInfo>>>();
    }

    void addColumnsInfo(Connection connection, String schemaName, String tableName, Map<String, ColumnEntity> alreadyLoadedColumns) throws SQLException {
        Map<String, List<ColumnInfo>> map = columnsInfo.get(schemaName);
        if (map == null) {
            map = new HashMap<String, List<ColumnInfo>>();
            columnsInfo.put(schemaName, map);
        }
        map.put(tableName, DataProvider.getAllColumns(connection, schemaName, tableName, alreadyLoadedColumns));
    }

    List<ColumnInfo> getColumnsInfo(String schemaName, String tableName) {
        return columnsInfo.get(schemaName).get(tableName);
    }
}

/**
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
public class TapSchemaHandler {

    private final Credentials credentials;
    private final String name;
    private boolean exists;

    private final List<String> allSchemas;
    private final Map<String, SchemaEntity> schemas;
    private final Map<String, List<KeyEntity>> schemaKeys;
    private final List<String> persisedSchemas;
    private final List<SchemaEntity> toRemoveSchemas;

    private final Map<String, List<String>> allSchemaTables;

    private final ColumnsInfo columnsInfo;

    private EntityManager entityManager;

    private Connection connection;

    public TapSchemaHandler(Credentials credentials, String tapSchemaName, boolean exists) throws SQLException {
        this.credentials = credentials;
        this.name = tapSchemaName;
        this.exists = exists;

        Connection conn = getConnection();

        try {
            this.allSchemas = DataProvider.getAllSchemasNames(conn);

            this.schemas = new HashMap<String, SchemaEntity>();
            this.schemaKeys = new HashMap<String, List<KeyEntity>>();
            this.persisedSchemas = new ArrayList<String>();
            this.toRemoveSchemas = new ArrayList<SchemaEntity>();

            this.allSchemaTables = new HashMap<String, List<String>>();

            this.columnsInfo = new ColumnsInfo();

            if (exists) {
                entityManager = DataProvider.getEntityManager(credentials, tapSchemaName);
                CriteriaBuilder cb = entityManager.getCriteriaBuilder();
                CriteriaQuery<SchemaEntity> cq = cb.createQuery(SchemaEntity.class);
                for (SchemaEntity schemaEntity : entityManager.createQuery(cq).getResultList()) {
                    String schemaName = schemaEntity.getName();
                    addSchema(schemaEntity);
                    addSchemaInfo(conn, schemaName);
                    persisedSchemas.add(schemaName);
                    for (TableEntity tableEntity : schemaEntity.getTables().values()) {
                        columnsInfo.addColumnsInfo(conn, schemaName, tableEntity.getTableName(), tableEntity.getColumns());
                    }
                }
            }

        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection();
        }
    }

    private void addSchemaInfo(Connection connection, String schemaName) throws SQLException {
        schemaKeys.put(schemaName, DataProvider.getSchemaKeys(connection, schemaName));
        allSchemaTables.put(schemaName, DataProvider.getAllTablesNames(connection, schemaName));
    }

    public SchemaEntity getNewSchema(String schemaName) throws SQLException {
        SchemaEntity schemaEntity = new SchemaEntity(schemaName);

        Connection conn = getConnection();
        try {
            addSchemaInfo(conn, schemaName);
        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection();
        }

        return schemaEntity;
    }

    public final void addSchema(SchemaEntity schemaEntity) {
        schemas.put(schemaEntity.getName(), schemaEntity);
    }

    public final void removeSchema(String schemaName) {
        if (persisedSchemas.contains(schemaName)) {
            toRemoveSchemas.add(schemas.get(schemaName));
            persisedSchemas.remove(schemaName);
        }
        schemas.remove(schemaName);
    }

    public Collection<SchemaEntity> getSchemas() {
        return schemas.values();
    }

    public List<String> getAllSchemas() {
        return allSchemas;
    }

    public List<String> getAllTables(String schemaName) {
        return allSchemaTables.get(schemaName);
    }

    public List<ColumnInfo> getColumnInfo(String schemaName, String tableName) {
        return columnsInfo.getColumnsInfo(schemaName, tableName);
    }

    private int getMaxKey() {
        int maxKey = 0;
        for (SchemaEntity schemaEntity : schemas.values()) {
            for (TableEntity tableEntity : schemaEntity.getTables().values()) {
                for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
                for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
            }
        }
        return maxKey;
    }

    public TableEntity getNewTable(String schemaName, String tableName) throws SQLException {

        SchemaEntity schema = schemas.get(schemaName);
        TableEntity table = new TableEntity(schemaName + "." + tableName);

        // Add table keys
        int maxKey = getMaxKey();
        for (KeyEntity keyEntity : schemaKeys.get(schemaName)) {
            String fromTable = keyEntity.getFromTableName();
            String targetTable = keyEntity.getTargetTableName();
            if (schema.getTables().containsKey(fromTable)
                    && schema.getTables().containsKey(targetTable)) {
                String key = maxKey + "";
                keyEntity.setKeyId(key);
                for (KeyColumnEntity keyColumn : keyEntity.getKeyColumns()) {
                    keyColumn.setKeyId(key);
                }

                schema.getTables().get(fromTable).getFromKeys().add(keyEntity);
                schema.getTables().get(targetTable).getTargetKeys().add(keyEntity);

                maxKey++;
            }
        }

        Connection conn = getConnection();
        try {
            // Add columns info
            // For performance reasons all possible column entities are loaded together
            columnsInfo.addColumnsInfo(conn, schemaName, tableName, null);
        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection();
        }

        return table;
    }

    public final Connection getConnection() throws SQLException {
        if (connection == null) {
            connection = credentials.getConnection();
        }
        return connection;
    }

    public final void closeConnection() throws SQLException {
        if (connection != null) {
            try {
                connection.close();
            } catch (SQLException e) {
                throw e;
            } finally {
                connection = null;
            }
        }
    }

    public void save() throws SQLException {
        if (!exists) {
            try {
                DataProvider.createDatabase(getConnection(), name);
            } catch (SQLException e) {
                throw e;
            } finally {
                closeConnection();
            }
            entityManager = DataProvider.getEntityManager(credentials, name);
            exists = true;
        }

        entityManager.getTransaction().begin();
        for (SchemaEntity toRemoveSchema : toRemoveSchemas) {
            entityManager.remove(toRemoveSchema);
        }
        for (SchemaEntity schemaEntity : schemas.values()) {
            entityManager.persist(schemaEntity);
        }
        entityManager.getTransaction().commit();
    }
}
